package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import java.awt.geom.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.*;

/**
 * A JPanel that displays a sequence of points as vertices, which
 * may optionally be dragged.
 */
abstract class LineComponent extends JPanel implements MouseListener, MouseMotionListener {

    private static final int MARGIN = 5;
    protected final PagePanel pagePanel;
    protected final Point2D[] points;
    protected final boolean[] readonly;
    protected final boolean closepath;

    protected Point dragPointScreen;
    private Shape targetArea;
    private int selectedVertex = -1;

    LineComponent(PagePanel pagePanel, Point2D[] points, boolean[] readonly, boolean close) {
        this.pagePanel = pagePanel;
        this.points = points;
        this.readonly = readonly;
        this.closepath = close;
        boolean anyadjustable = false;
        for (int i=0;!anyadjustable && i<readonly.length;i++) {
            anyadjustable |= !readonly[i];
        }
        if (anyadjustable) {
            addMouseListener(this);
            addMouseMotionListener(this);
        }
        setBorder(AnnotationComponentFactory.FOCUSBORDER);
        setOpaque(false);
    }

    /**
     * Called when a popup trigger is received
     */
    abstract void showPopup(Point point);

    /**
     * Called when the annotation needs to be redrawn.
     * @param rebuild true if the points have been altered, false
     * if we just need to call PagePanel.redrawAnnotaton()
     */
    abstract void redrawAnnotation(boolean rebuild);

    private Point getScreenPoint(int point) {
        return (Point)pagePanel.getPageToScreenTransform().transform(points[point], new Point());
    }

    public void mouseClicked(MouseEvent e) {
        if (e.isPopupTrigger()) {
            showPopup(e.getPoint());
            reset();
            redrawAnnotation(false);
            repaint();
        }
    }

    private void reset() {
        selectedVertex = -1;
        dragPointScreen = null;
        targetArea = null;
        setCursor(null);
    }

    public void mousePressed(MouseEvent e) {
        if (e.isPopupTrigger()) {
            mouseClicked(e);
            return;
        }
        boolean caught = false;
        Point mousepoint = SwingUtilities.convertPoint(this, e.getPoint(), pagePanel);
        for (int i=0;!caught && i < points.length;i++) {
            Point screenpoint = getScreenPoint(i);
            if (!readonly[i] && mousepoint.distance(screenpoint) < MARGIN) {
                selectedVertex = i;
                caught = true;
            }
        }
        if (caught) {
            dragPointScreen = mousepoint;
            Rectangle pageBounds = getParent().getBounds();
            putClientProperty("images", null);
            setBounds(0, 0, pageBounds.width, pageBounds.height);
            repaint();
            new Thread() {
                public void run() {
                    try {
                        while (dragPointScreen!=null) {
                            Thread.sleep(AbstractRegionSelector.CRAWLSPEED);
                            repaint();
                        }
                    } catch (InterruptedException e) { }
                }
            }.start();
        } else {
            dispatchMouseEvent(e);
        }
    }

    private void dispatchMouseEvent(MouseEvent e) {
        // TODO - currently only sends to parents, we really want to
        // send to next item in the stacking hierarchy in case two
        // components overlap
        Point offset = getLocation();
        e.translatePoint(offset.x, offset.y);
        getParent().dispatchEvent(e);
    }

    public void mouseReleased(MouseEvent e) {
        if (e.isPopupTrigger()) {
            mouseClicked(e);
            return;
        }
        if (selectedVertex != -1) {
            Point screenpoint = SwingUtilities.convertPoint(this, e.getPoint(), pagePanel);
            if (isAllowedDragPoint(screenpoint)) {
                Point2D pagepoint = pagePanel.getScreenToPageTransform().transform(screenpoint, null);
                points[selectedVertex] = pagepoint;
            }
            redrawAnnotation(true);
            reset();
            repaint();
        }
        dispatchMouseEvent(e);
    }

    public void mouseEntered(MouseEvent e) {
        mouseMoved(e);
    }

    public void mouseExited(MouseEvent e) {
        if (targetArea != null) {
            targetArea = null;
            setCursor(null);
            repaint();
        }
    }

    public void mouseDragged(MouseEvent e) {
        if (selectedVertex > -1) {
            // Ensure drag point stays inside page bounds
            Point p = SwingUtilities.convertPoint(this, e.getPoint(), pagePanel);
            Rectangle b = getBounds();
            p.x = Math.max(b.x, Math.min(b.width-1, p.x));
            p.y = Math.max(b.y, Math.min(b.height-1, p.y));
            if (isAllowedDragPoint(p)) {
                dragPointScreen = p;
            }
            repaint();
         } else {
             dispatchMouseEvent(e);
         }
    }

    protected boolean isAllowedDragPoint(Point p) {
        return true;
    }

    public void mouseMoved(MouseEvent e) {
        Point mousepoint = SwingUtilities.convertPoint(this, e.getPoint(), pagePanel);
        Point p = null;
        for (int i=0;p==null && i < points.length;i++) {
            Point screenpoint = getScreenPoint(i);
            if (!readonly[i] && mousepoint.distance(screenpoint) < MARGIN) {
                p = SwingUtilities.convertPoint(pagePanel, screenpoint, this);
            }
        }
        if (p!=null) {
            targetArea = new Ellipse2D.Float(p.x - MARGIN, p.y - MARGIN, MARGIN * 2, MARGIN * 2);
            setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
            repaint();
        } else if (targetArea != null) {
            targetArea = null;
            setCursor(null);
            repaint();
        }
        if (targetArea==null) {
             dispatchMouseEvent(e);
        }
    }

    public void paintComponent(Graphics g) {
        if (dragPointScreen != null) {
            g.setColor(Color.black);
            BasicStroke stroke = new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 15, new float[] { AbstractRegionSelector.CRAWLLENGTH, AbstractRegionSelector.CRAWLLENGTH }, (int)((System.currentTimeMillis()/AbstractRegionSelector.CRAWLSPEED)%(AbstractRegionSelector.CRAWLLENGTH*2)));
            ((Graphics2D)g).setStroke(stroke);

            Point firstp = null, lastp = null;
            for (int i=0;i < points.length;i++) {
                Point p = i==selectedVertex ? dragPointScreen : getScreenPoint(i);
                p = SwingUtilities.convertPoint(pagePanel, p, this);
                if (i > 0) {
                    g.drawLine(lastp.x, lastp.y, p.x, p.y);
                }
                lastp = p;
                if (firstp==null) {
                    firstp = p;
                }
            }
            if (closepath) {
                g.drawLine(lastp.x, lastp.y, firstp.x, firstp.y);
            }
        } else {
            AnnotationComponentFactory.paintComponent(this, this.ui, g);
            if (targetArea != null) {
                g.setColor(new Color(0xA0FF0000, true));
                ((Graphics2D)g).fill(targetArea);
            }
        }
    }
}
